/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Sun Feb 24 16:10:53 1991 by burrows        */
#include <stdio.h>
#include <ctype.h>
#include <strings.h>
#include "indexserver.h"

#define HASHSIZE	17
#define ALIGN (sizeof (char *))
#define BLOCKSIZE	4096

struct blocklist {
	struct blocklist *next;
	int size;
	int rest;
	char data[1];
};

struct stringsave {
	struct blocklist *blocks;
	struct strlist *hash[HASHSIZE];
};

struct strlist {
	struct strlist *next;
	int len;
	char str[1];
};

static char *parsesubj ();
static char *parseid ();
static char *savestr ();
static void initsave ();
static void freesave ();
static struct msgset *findothers ();

/* ARGSUSED */
struct msgset *discussfunc (index2, groups, fulldomain, domain,
	arg, arglen, not, perr)
struct index2 *index2;
struct groups *groups;
struct msgset *fulldomain;
struct msgset *domain;
char *arg;
int arglen;
int not;
int *perr;
{
	int c;
	struct msgset *stilltoscan, *newtoscan, *tmp, *result, *subdomain;
	struct stringsave subj, mid;
	register long *x, l;
	register int i, j, k;
	long msg;

	c = nc (&arg);
	result = submsgset (index2, groups, fulldomain, fulldomain,
		nc, &arg, &c, 1, perr);

	if (result == 0)
		return (result);
	
	initsave (&subj);
	initsave (&mid);

	stilltoscan = ormsgset (result, (struct msgset *) 0);
	subdomain = notmsgset (fulldomain, result);

	while ((i = countmsgset (stilltoscan)) > 0 && i < 100) {
		newtoscan = 0;
		x = stilltoscan->bitset;
		for (i = 0, j = stilltoscan->nbitset; i < j; i++) {
			if ((l = x[i]) != 0) {
				for (k = 0; k < LONGBITS; k++) {
					if (l & (1 << k)) {
						msg = (long) k+i*LONGBITS;
						newtoscan = findothers (index2,
						    subdomain, parsesubj, msg,
						    SUBJECTFIELD, &subj,
						    newtoscan,samesubjselect,1);
						newtoscan = findothers (index2,
						    subdomain, parseid, msg,
						    INREPLYTOFIELD, &mid,
						    newtoscan, idselect, 0);
						newtoscan = findothers (index2,
						    subdomain, parseid, msg,
						    MESSAGEIDFIELD, &mid,
						    newtoscan, idselect, 0);
						newtoscan = findothers (index2,
						    subdomain, parseid, msg,
						    REFERENCESFIELD, &mid,
						    newtoscan, idselect, 0);
					}
				}
			}
		}
		subdomain = into_notmsgset (subdomain, newtoscan);
		result = into_ormsgset (result, newtoscan);
		freemsgset (stilltoscan);
		stilltoscan = newtoscan;
	}

	freesave (&subj);
	freesave (&mid);
	freemsgset (stilltoscan);
	freemsgset (subdomain);

	if (i != 0) {
		*perr = 1;
		(void) printf ("?sorry: discussion too big\n");
		freemsgset (result);
		return (0);
	}
	if (not) {
		tmp = notmsgset (domain, result);
		freemsgset (result);
		result = tmp;
	} else {
		result = into_andmsgset (result, domain);
	}

	return (result);
}

static struct msgset *findothers (index2, domain, parse, msg,
	fnum, data, new, sel, literal)
struct index2 *index2;
struct msgset *domain;
char *(*parse) ();
long msg;
int fnum;
struct stringsave *data;
register struct msgset *new;
int (*sel) ();
int literal;
{
	register char *f, *saved;
	int len;
	register struct msgset *tmp;

	f = field (index2, msg, fnum);

	do {
		f = (*parse) (f, &len);
		if (len > 0 && len < 128) {
			saved = savestr (data, f, len);
			if (saved != 0) {
				tmp = filterselectmsg (index2, domain,
					sel, saved, 0, literal);
				new = into_ormsgset (new, tmp);
				freemsgset (tmp);
			}
		}
		f += len;
	} while (len > 0 && *f != '\0');

	return (new);
}

static char *parsesubj (s, plen)
register char *s;
register int *plen;
{
	for (;;) {
		while (isspace (*s & 0xff))
			s++;
		if ((*s == 'r' || *s == 'R') && (s[1] == 'e' || s[1] == 'E') &&
		    (s[2] == ';' || s[2] == ':' || isspace (s[2] & 0xff)))
			s += 3;
		else
			break;
	}

	*plen = strlen (s);

	return (s);
}

static char *parseid (s, plen)
register char *s;
register int *plen;
{
	register char *p;
	register int c;
	register char *k;

again:
	while (isspace (*s & 0xff))
		s++;
	p = s;
	while ((c = *p++) != 0 && c != '<' && (c = *p++) != 0 && c != '<' &&
	       (c = *p++) != 0 && c != '<' && (c = *p++) != 0 && c != '<')
		;

	if (c != '<') {
		*plen = 0;
		return (s);
	}

	s = p;
	while ((c = *p++) != 0 && c!='>' && (c = *p++) != 0 && c!='>' &&
	       (c = *p++) != 0 && c!='>' && (c = *p++) != 0 && c!='>')
		;
	if (c != '>') {
		*plen = 0;
		return (s);
	}
		/* following hack is to reject short, all alphabetic ids,
		 * such as <muller>
		 */
	if (p - s < 10) {
		for (k = s; *k != '>'; k++)
			if (!isalpha (*k))
				break;
		if (*k == '>') {
			s = p;
			goto again;
		}
	}
	*plen = (p - s) - 1;
	return (s);
}


static void initsave (data)
struct stringsave *data;
{
	bzero ((char *) data, sizeof (*data));
}

static int blockcnt = 0;

static char *savestr (data, str, len)
struct stringsave *data;
register char *str;
register int len;
{
	register int hash = len;
	register struct strlist *p;
	int size;
	struct blocklist *block;

	if (len > 16)
		hash ^= str[16];
	hash ^= *str;
	hash %= HASHSIZE;

	for (p = data->hash[hash]; p != 0; p = p->next) {
		if (len == p->len && bcmp (str, p->str, len) == 0)
			return (0);
	}
	
	size = sizeof (struct strlist) + len + 1;
	size = (size + ALIGN - 1) & ~(ALIGN - 1);

	block = data->blocks;
	if (block == 0 || size > block->size - block->rest) {
		block = (struct blocklist *) malloc (sizeof (struct blocklist) +
			BLOCKSIZE);
		if (block == 0) {
			perror ("indexserver: malloc");
			exit (2);
		}
/* fprintf (stderr, "blockcnt=%d\n", ++blockcnt); */
		block->next = data->blocks;
		data->blocks = block;
		block->size = BLOCKSIZE;
		block->rest = 0;
	}

	p = (struct strlist *) &block->data[block->rest];
	block->rest += size;
	p->len = len;
	bcopy (str, p->str, len);
	p->str[len] = '\0';

	p->next = data->hash[hash];
	data->hash[hash] = p;

	return (p->str);
}

static void freesave (data)
struct stringsave *data;
{
	struct blocklist *x, *y;

	for (x = data->blocks; x != 0; x = y) {
		y = x->next;
		free ((char *) x);
/* fprintf (stderr, "blockcnt=%d\n", --blockcnt); */
	}
}
