/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Thu Feb 28 13:24:12 1991 by burrows        */
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <strings.h>
#include "upperlower.h"
#include "avl.h"

static long crunch ();

static int wordstart[MAXCHAR+1];
static int inword[MAXCHAR+1];
static int printword ();
static void addword ();
extern char *emalloc ();

extern long atol ();
extern long lseek ();
extern void exit ();
extern void abort ();
extern void perror ();
extern void bcopy ();
extern char *malloc ();

#define HASHSIZE 4099

struct state {
	FILE *ofp;
	FILE *ifp;
	long count;
	long pcount;
	char buffer[100];
};

static void buildfrombinary ();

static void usage ()
{
	(void) fprintf (stderr, "usage: maketextindex [-m] [-i] [-u uid] [-p pos] [-o output] [-t] [file...]\n");
	exit (2);
}

main (argc, argv)
int argc;
char *argv[];
{
	register int i;
	long pos = 0L;
	long uid = 0L;
	char *output = 0;
	char *tmp = 0;
	char ifile[1024];
	char tfile[1024];
	int fd;
	int argn;
	FILE *ofp, *ifp;
	struct avl *tree = 0;
	struct wordlist *hashtable[HASHSIZE];
	struct state state;
	int merge = 0;
	int ignorecase = 0;
	char name[1024];
	int uidandname = 0;

#ifdef SETLOCALE
	(void) setlocale (0, "");
#endif

	argn = 1;
	while (argn < argc && argv[argn][0] == '-' && argv[argn][1] != '\0') {
		switch (argv[argn][1])
		{
		case 'p': argn++;
			  if (argn >= argc) usage ();
			  pos = atol (argv[argn]);
			  break;
		case 'u': argn++;
			  if (argn >= argc) usage ();
			  uid = atol (argv[argn]);
			  break;
		case 'm': merge = 1;
			  break;
		case 'o': argn++;
			  if (argn >= argc) usage ();
			  output = argv[argn];
			  break;
		case 'U': uidandname = 1;
			  break;
		case 't': tmp = tfile;
			  break;
		case 'i': ignorecase = 1;
			  break;
		default:  usage ();
		}
		argn++;
	}

	if (output == 0)
		tmp = 0;
	else {
		(void) strcpy (ifile, output);
		(void) strcat (ifile, ".i");
		(void) strcpy (tfile, output);
		(void) strcat (tfile, ".t");
	}

	for (i = 0; i <= MAXCHAR; i++) {
		wordstart[i] = isalnum (i);
		inword[i] = isalnum (i);
	}
	wordstart[1] = 1;
	initupperlower ();

	bzero ((char *) hashtable, sizeof (hashtable));

	if (output != 0)
		ofp = fopen (tmp?tmp:output, "w");
	else
		ofp = stdout;
	
	if (ofp == NULL) {
		perror (tmp?tmp:output);
		exit (3);
	}

	ifp = NULL;
	if (output != 0) {
		if ((ifp = fopen (ifile, "w")) == NULL) {
			perror (ifile);
			(void) unlink (tmp?tmp:output);
			exit (4);
		}
	}

	if (merge && !uidandname)
		buildfrombinary (stdin, &tree, hashtable);

	if (argn >= argc && !merge && !uidandname) {
		if (pos != 0)
			(void) lseek (0, pos, 0);
		uid = crunch (0, uid, &tree, hashtable, ignorecase);
	}

	while (argn < argc) {
		if (strcmp (argv[argn], "-") == 0 && !merge && !uidandname)
			fd = 0;
		else
			fd = open (argv[argn], 0);
		if (fd >= 0) {
			if (pos != 0) {
				(void) lseek (fd, pos, 0);
				pos = 0L;
			}
			uid = crunch (fd, uid, &tree, hashtable, ignorecase);
			uid++;
			if (fd > 0)
				(void) close (fd);
		} else {
			perror (argv[argn]);
			exit (5);
		}
		argn++;
	}

	if (uidandname) {
		while (scanf ("%d %s", &uid, name) == 2) {
			fd = open (name, 0);
			if (fd >= 0) {
				(void) crunch (fd, uid, &tree,
					hashtable, ignorecase);
				(void) close (fd);
			} else {
				perror (name);
				/* exit (5); */
			}
		}
		if (ferror (stdin) || !feof (stdin)) {
			(void) fprintf (stderr,
				"maketextindex: error on stdin\n");
			exit (5);
		}
	}

	state.buffer[0] = '\0';
	state.ofp = ofp;
	state.ifp = ifp;
	state.pcount = -100000L;
	state.count = 0L;
	(void) avl_iterate (&tree, printword, (char *) &state);

	(void) fflush (ofp);
	if (ferror (ofp) || (fsync (fileno (ofp)) < 0 && errno != EINVAL)) {
		if (output) {
			perror (tmp?tmp:output);
			(void) unlink (tmp?tmp:output);
			(void) unlink (ifile);
		} else
			perror ("stdout");
		exit (6);
	}

	if (ifp != NULL) {
		(void) fflush (ifp);
		if (ferror (ifp)) {
			perror (ifile);
			(void) unlink (tmp?tmp:output);
			(void) unlink (ifile);
			exit (7);
		}
	}

	if (tmp != 0) {
		if (rename (tmp, output) < 0) {
			perror (tmp);
			(void) unlink (tmp);
			(void) unlink (ifile);
			exit (8);
		}
	}

	exit (0);
}

/* -------------------------------------------------------------- */
#define NUIDREADY 1000
static struct uidlist *uidready;
static int uidcount = 0;

static void mkuidready ()
{
	uidcount = NUIDREADY;
	uidready = (struct uidlist *) emalloc (sizeof (struct uidlist) * uidcount);
}

#define NEWUID(x)	do { if (uidcount==0) mkuidready (); x = uidready++; uidcount--; } while (0)
/* -------------------------------------------------------------- */
#define NWORDREADY 10000
static char *wordready;
static int wordcount = 0;

static void mkwordready ()
{
	wordcount = NWORDREADY;
	wordready = emalloc (wordcount);
}

#define ALIGN (sizeof (char *))
#define NEWWORD(word,n) do { register int a = (sizeof (struct wordlist)+n+ALIGN+1) & ~(ALIGN-1); if (wordcount < a) mkwordready (); word = (struct wordlist *) wordready; wordready += a; wordcount -= a; } while (0)

/* -------------------------------------------------------------- */

#define BSIZ 8192

static long crunch (fd, uid, ptree, hashtable, ignorecase)
int fd;
long uid;
struct avl **ptree;
struct wordlist *hashtable[];
int ignorecase;
{
	char buffer[BSIZ+512];
	int nread = 1;
	int got = 0;
	int toread = BSIZ;
	register char *p, *word;
	char *end;

	while ((nread > 0 && (nread = read (fd, buffer+got, toread)) > 0) ||
				got > 0) {
		if (nread < 0) {
			perror ("read");
			exit (2);
		}
		end = &buffer[got+nread];
		*end = '\n';
		end[1] = 'a';
		end[2] = '\n';
		if (nread == 0)
			end++;
		
		word = buffer;
		do {
			{
				register int *ws = wordstart;

				while (!ws[*word++ & MAXCHAR] &&
				       !ws[*word++ & MAXCHAR] &&
				       !ws[*word++ & MAXCHAR] &&
				       !ws[*word++ & MAXCHAR])
					;
			}
			word--;
			if (*word == '\001') {
				if (nread > 0 && end-word < 10)
					break;
				if (end-word >= 10) {
					if (bcmp ("\1\1\1\1\n\1\1\1\1\n",
						  word, 10) == 0) {
						uid++;
						word += 10;
						p = word;
						continue;
					}
				}
				word++;
				p = word;
				continue;
			}

			p = word;
			if (!ignorecase) {
				register int *iw = inword;

				while (iw[*p++ & MAXCHAR] &&
				       iw[*p++ & MAXCHAR] &&
				       iw[*p++ & MAXCHAR] &&
				       iw[*p++ & MAXCHAR])
					;
				p--;
			} else {
				register char *l = makelower;
				register int c;
				register int *iw = inword;

				while (iw[c = (*p & MAXCHAR)])
					*p++ = l[c];
			}

			if (p < end) {
				addword (word, p, uid, ptree, hashtable);
				word = p;
			}
		} while (p < end);
		got = end-word;
		if (got <= 0)
			got = 0;
		else
			bcopy (word, buffer, got);
		toread = sizeof (buffer) - got - 10;
		if (toread > BSIZ)
			toread = BSIZ;
	}
	if (nread < 0) {
		perror ("read");
		exit (2);
	}

	return (uid);
}

static int printword (state, w)
struct state *state;
struct wordlist **w;
{
	struct uidlist *u, *e;
	register char *p = state->buffer;
	register char *q = (*w)->word;
	register int i;
	long ln = 0;
	register FILE *ofp = state->ofp;
	register long count = state->count;

	if (state->ifp != NULL && count > state->pcount + 4096)
		(void) fprintf (state->ifp, "%s %ld\n", q, state->pcount=count);

	while (*q == *p && *p != '\0')
		p++, q++;
	
	i = p - state->buffer;
	putc (0x80 | i, ofp);
	count += 2-i;
	while (i < 0x3f && *q != '\0')
		putc (*p++ = *q++, ofp), i++;
	*p = '\0';
	while (*q != '\0')
		putc (*q++, ofp);
	putc ('\0', ofp);
	
	count += (q - (*w)->word);

	e = u = (*w)->uids;

	if (u != 0) {
		do {
			u = u->next;
			i = u->uid - ln;
			ln = u->uid;
			while (i > 0x7f) {
				putc (0xc0 | i, ofp);
				i >>= 6;
				count++;
			}
			putc (i, ofp);
			count++;
		} while (u != e);
	}

	state->count = count;

	return (0);
}

char *emalloc (n)
int n;
{
	char *p = malloc ((unsigned) n);

	if (p == 0) {
		perror ("malloc");
		exit (2);
	}
	return (p);
}

static void buildfrombinary (ifp, ptree, hashtable)
register FILE *ifp;
struct avl **ptree;
struct wordlist *hashtable[];
{
	char buf[2000];
	register int c;
	register char *s;
	register long n, ln;
	int shift;

	ln = 0;

	c = getc (ifp);

	while ((c & 0xc0) == 0x80) {
		c &= 0x3f;
		s = &buf[c];

		c = getc (ifp);
		
		while (c != '\0' && c != EOF) {
			*s++ = c;
			c = getc (ifp);
		}
		if (c != EOF)
			c = getc (ifp);

		*s = '\1';

		ln = 0;

		while ((c & 0xc0) != 0x80 && c != EOF) { 
			n = 0;
			shift = 0;
			while ((c & 0xc0) == 0xc0 && c != EOF) {
				n |= (c & 0x3f) << shift;
				shift += 6;
				c = getc (ifp);
			}
			n |= (c & 0x7f) << shift;
			n += ln;
			ln = n;
			addword (buf, s, ln, ptree, hashtable);
			c = getc (ifp);
		}
	}
}

static void addword (word, end, uid, ptree, hashtable)
register char *word;
register char *end;
long uid;
struct avl **ptree;
struct wordlist *hashtable[];
{
	register struct wordlist *w;
	struct wordlist *xw, **pw;
	int hash;

	{
		register int x, y = end - word;

		x = (*word & 0xff) + y +
			((end[-1] & 0xff) << 8);
		if (y > 2)
			x += (word[2] & 0xff) << 16;
		x %= HASHSIZE;
		w = hashtable[x];
		hash = x;
	}

	while (w != 0) {
		register char *q = w->word;
		register char *s = word;
		
		if (*end != '\0') {
			while (*q++ == *s++ &&
			       *q++ == *s++ &&
			       *q++ == *s++ &&
			       *q++ == *s++ &&
			       *q++ == *s++)
				;
			q--, s--;
		} else {
			while (*q == *s && s < end)
				q++, s++;
		}
		if (s == end && *q == '\0')
			break;
		w = w->hashchain;
	}

	if (w == 0) {
		NEWWORD (w, end-word);
		w->uids = 0;
		w->hashchain = hashtable[hash];
		hashtable[hash] = w;

		bcopy (word, w->word, end-word);
		w->word[end-word] = '\0';
		xw = w;
		pw = avl_add (ptree, strcmp, &xw);
		if (*pw != w)
			abort ();
	}

	if (w->uids == 0 || w->uids->uid != uid) {
		register struct uidlist *u;

		NEWUID (u);
		u->uid = uid;
		if (w->uids == 0)
			w->uids = u->next = u;
		else {
			u->next = w->uids->next;
			w->uids->next = u;
			w->uids = u;
		}
	}
}
